/*
 *  optimise.c
 *  Numerical Fibre Generator
 *
 *  Created by Tom Close on 25/06/08.
 *  Copyright 2008 Tom Close.
 *  Distributed under the GNU General Public Licence.
 *
 *
 *
 *  This file is part of 'Numerical Fibre Generator'.
 *
 *  'Numerical Fibre Generator' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  'Numerical Fibre Generator' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'Numerical Fibre Generator'.  If not, see <http://www.gnu.org/licenses/>
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <gsl/gsl_multimin.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "shared.h"
#include "cost_function.h"

#include "optimise.h"

#define COST_TERM_WEIGHTS_DEFAULT {1.0e5, 1.0, 20.0}
#define COST_TERM_POWERS_DEFAULT {2.0, 2.0, 2.0}
#define SAMPLE_DENSITY_DEFAULT 100.0
#define GRID_REFERENCE_SIZE_DEFAULT 10
#define INITIAL_STEP_SIZE_DEFAULT 1.0
#define STEP_TOL_DEFAULT  0.1
#define OPTIMIZATION_TOL_DEFAULT 1.0e-2
#define MAX_ITERATIONS_DEFAULT 100
/*#define MAX_NUM_COST_FUNCTION_CALLS_DEFAULT 30*/
#define ITERATIONS_SAVE_FREQ_DEFAULT 0
#define JITTER_STDEV_DEFAULT 0.01
#define JITTER_SEED_DEFAULT 1193125662
#define FOV_DEFAULT 1.2




/* Handles the input/output and variable initialisation.  All the 'work' is performed by 'rand_init_collection'*/
int main(int argc, char **argv) {

	int error;
	double sample_density, initial_step_size, step_tol, optimization_tol, fov, jitter_dev;
	int grid_reference_size, max_iterations, iterations_save_freq; 
	unsigned int jitter_seed;


	Strand_collection c;
	DIR *dir;	
	FILE *param_file, *file;
	char *param_path, *input_dir_path,  *output_dir_path, *iterations_save_path, *path;
	int stage_num;

	int num_params, num_lines;
	char line[200], key[200];
	double value;

	double cost_term_weights[3] = COST_TERM_WEIGHTS_DEFAULT;
	double cost_term_powers[3] = COST_TERM_POWERS_DEFAULT;
	sample_density = SAMPLE_DENSITY_DEFAULT;
	grid_reference_size = GRID_REFERENCE_SIZE_DEFAULT;
	initial_step_size = INITIAL_STEP_SIZE_DEFAULT;
	step_tol = STEP_TOL_DEFAULT;
	optimization_tol = OPTIMIZATION_TOL_DEFAULT;
	max_iterations = MAX_ITERATIONS_DEFAULT;
/*	max_num_cost_function_calls = MAX_NUM_COST_FUNCTION_CALLS_DEFAULT;*/
	jitter_dev = JITTER_STDEV_DEFAULT;
	jitter_seed = JITTER_SEED_DEFAULT;
	iterations_save_freq = ITERATIONS_SAVE_FREQ_DEFAULT;
	fov = FOV_DEFAULT;

	param_path = "";


	if (argc <= 2 || argc >= 5) {
		printf("\nNumerical Fibre Generator (%s): 'optimise'\n\nNumber of suplied arguments %d.\n\n Usage: \n\t arg[1]: input directory \n\t arg[2]: output directory \n\t arg[3]: parameters file\n\n", VERSION_NUM, (argc-1));
		exit(EXIT_FAILURE);
		

	} else if (argc == 4) {
	
		param_path = argv[3];
		

		if ((param_file = fopen(param_path, "r")) == NULL) {
			printf("Error! Could not open file %s\n", param_path);
			error = 1;
		}

			
		num_params = 0;
		num_lines = 0;
		
		printf("\n\nReading parameters file: %s\n\n", param_path);
		fflush(stdout);
		
		while (fgets(line, 200, param_file) != NULL) {
			
			if (sscanf(line, "%s %lf", key, &value) != EOF) {
				
				if (!strcmp(key, "fov")) {
					fov = value;
					printf("Read value for fov, %g\n", value);
					
				} else if (!strcmp(key, "cost_repulsion_weight")) {
					cost_term_weights[0] = value;
					printf("Read value for cost_repulsion_weight, %g\n", value);
				
				}  else if (!strcmp(key, "cost_bending_weight")) {
					cost_term_weights[1] = value;
					printf("Read value for cost_bending_weight, %g\n", value);
				
				}  else if (!strcmp(key, "cost_length_weight")) {
					cost_term_weights[2] = value;
					printf("Read value for cost_length_weight, %g\n", value);
				
				} else if (!strcmp(key, "cost_repulsion_power")) {
					cost_term_powers[0] = value;
					printf("Read value for cost_repulsion_power, %g\n", value);
				
				} else if (!strcmp(key, "cost_bending_power")) {
					cost_term_powers[1] = value;
					printf("Read value for cost_bending_power, %g\n", value);
				
				} else if (!strcmp(key, "cost_length_power")) {
					cost_term_powers[2] = value;
					printf("Read value for cost_length_power, %g\n", value);
				
				}  else if (!strcmp(key, "sample_density")) {
					sample_density = value;
					printf("Read value for sample_density, %g\n", value);
				
				} else if (!strcmp(key, "grid_reference_size")) {
					grid_reference_size = value;
					printf("Read value for grid_reference_size, %g\n", value);
				
				} else if (!strcmp(key, "initial_step_size")) {
					initial_step_size = value;
					printf("Read value for initial_step_size, %g\n", value);
				
				} else if (!strcmp(key, "step_tol")) {
					step_tol = value;
					printf("Read value for step_tol, %g\n", value);
				
				} else if (!strcmp(key, "optimization_tol")) {
					optimization_tol = value;
					printf("Read value for optimization_tol, %g\n", value);
				
				} else if (!strcmp(key, "jitter_dev")) {
					jitter_dev = value;
					printf("Read value for jitter_dev, %g\n", value);
				
				} else if (!strcmp(key, "jitter_seed")) {
					jitter_seed = (unsigned int)value;
					printf("Read value for jitter_seed, %u\n", jitter_seed);
				
				} else if (!strcmp(key, "max_iterations")) {
					max_iterations = (int)value;
					printf("Read value for max_iterations, %d\n", max_iterations);
				
				} /*else if (!strcmp(key, "max_num_cost_function_calls")) {
					max_num_cost_function_calls = (int)value;
					printf("Read value for max_num_cost_function_calls, %d\n", max_num_cost_function_calls);
				
				} */ else if (!strcmp(key, "iterations_save_freq")) {
					
					iterations_save_freq = (int)value;
					printf("Read value for iterations_save_freq, %d\n", iterations_save_freq);
					
				} else {
					num_params--;
				}
				fflush(stdout);
				num_params++;
			}
			
			num_lines++;
		}
		
		printf("\nRead in %d (of possible 16) key/value pairs\n", num_params); 

	}
	
	fflush(stdout);
	
	
	input_dir_path = argv[1];
	output_dir_path = argv[2];
	
	if (load_collection(&c, input_dir_path)) {
		printf("Could not load strands from directory %s, exiting ...\n", input_dir_path);
		exit(EXIT_FAILURE);
	}
	
	/* If the fov supplied is greater than the furthest extent of the loaded strand then use it otherwise use the furthest extent */
	if (fov > c.fov) {
		c.fov = fov;	
	} 
	
	iterations_save_path = malloc(sizeof(char) * (strlen(output_dir_path) + 20)); 
	
	if (iterations_save_freq != 0) {

		if ((dir = opendir(output_dir_path)) == NULL) {
			if ((error = mkdir(output_dir_path, MY_PERMS)) > 0) {
				printf("Error! Could not create directory %s, (Error code: %d)!\n", output_dir_path, error);
				return 1;
			}
		} else {
			closedir(dir);
		}	
		
		
		
		sprintf(iterations_save_path, "%s%citerations", output_dir_path, DIR_SEP);

		if ((dir = opendir(iterations_save_path)) == NULL) {
			if ((error = mkdir(iterations_save_path, MY_PERMS)) > 0) {
				printf("Error! Could not create directory %s, (Error code: %d)!\n", argv[0], error);
				exit(EXIT_FAILURE);
			}
		} else {
			closedir(dir);
		}	
	}
	
	
	
	
	printf("\nOptimising strand control points...\n");	
	fflush(stdout);

	/* The optimisatioin algorithm */
	
	optimise(&c, cost_term_weights, cost_term_powers, sample_density, grid_reference_size, initial_step_size, step_tol, optimization_tol, jitter_dev, jitter_seed, max_iterations, iterations_save_freq, iterations_save_path);
	
	
	printf("\nWriting %d strand files to: %s...\n", c.num_strands, output_dir_path);
	fflush(stdout);
	
	
	if (save_collection(&c, output_dir_path, 0)) {
		printf("\nCould not save strands to directory %s, exiting ...\n", output_dir_path);
		exit(EXIT_FAILURE);
	
	}
	
	
	stage_num = copy_parameters(input_dir_path, output_dir_path);
	
	if (stage_num != -1) {
		path = malloc(sizeof(char) * (strlen(output_dir_path) + 80));	

		sprintf(path, "%s%cparameters%c%d_optimise_param.txt", output_dir_path, DIR_SEP, DIR_SEP, stage_num);
		
		if ( (file =fopen(path,"w")) == NULL) {
			printf("\nCould not open parameters save file %s.!\n", path);
		} 
										
		fprintf(file, "fov %g\n", fov);
		
		fprintf(file, "cost_repulsion_weight %g\n", cost_term_weights[0]);
		fprintf(file, "cost_bending_weight %g\n", cost_term_weights[1]);
		fprintf(file, "cost_length_weight %g\n", cost_term_weights[2]);	
		
		fprintf(file, "cost_repulsion_power %g\n", cost_term_powers[0]);	
		fprintf(file, "cost_bending_power %g\n", cost_term_powers[1]);
		fprintf(file, "cost_length_power %g\n", cost_term_powers[2]);
		
		fprintf(file, "sample_density %g\n", sample_density);
		fprintf(file, "grid_reference_size %d\n", grid_reference_size);
		fprintf(file, "initial_step_size %g\n", initial_step_size);
		fprintf(file, "step_tol %g\n", step_tol);
		fprintf(file, "optimization_tol %g\n", optimization_tol);
		fprintf(file, "jitter_dev %g\n", jitter_dev);
		fprintf(file, "jitter_seed %u\n", jitter_seed);
		fprintf(file, "max_iterations %d\n", max_iterations);
		fprintf(file, "iterations_save_freq %d\n", iterations_save_freq);
		
		
		if (fclose(file)) {
			printf("Error closing parameters file copy\n");
			exit(EXIT_FAILURE);
		}
	
		free(path);
	}
	
	free(iterations_save_path);
	

	exit(EXIT_SUCCESS);

}



/* Initialisation and iteration of optimisation algorithm */
void optimise(Strand_collection *c, double cost_term_weights[3], double cost_term_powers[3], double sample_density, int grid_reference_size, double initial_step_size, double step_tol, double optimization_tol, double jitter_dev, unsigned int jitter_seed, int max_iterations, int iterations_save_freq, char *iterations_save_path) {



	gsl_multimin_fdfminimizer *optimizer;
	const gsl_multimin_fdfminimizer_type *optimizer_type;
	gsl_multimin_function_fdf cost_function_struct;
	const gsl_vector *state_vector;
	Cost_function_params cost_function_params;
	
	clock_t start, stop;
	double time_for_iteration, total_time, total_time_min;
	
	size_t num_states;
	int iteration, status;
	double max_dx, dx, dx_sum, dx_sum_sq, dx_av, dx_2norm;
	
	unsigned int state_i;
	
	char *iterations_next_save_path, *iterations_next_dirname;
	
	int max_iter_label_width;

	
	gsl_rng *rand_gen;
	
	Sample_block *samples_store; 
	Reference_block *grid_reference;
	
	max_iter_label_width = num_decimal_places(max_iterations);
	
	iterations_next_save_path = malloc(sizeof(char) * (strlen(iterations_save_path) + 20 + max_iter_label_width));
	
	/*num_cost_function_calls = 0;	*/
				
	/* Allocate memory structures, which will be reused between successive iterations.  Prevents the algorithm from repetitively asking the OS for large blocks of memory, which can become a problem for very large numbers of strands (hence large memory blocks) */
	    					
	allocate_storage_blocks(&samples_store, &grid_reference, grid_reference_size);	
		
	/* Initialise input parameters of the GSL optimiser (implementing the L-BFGS algorithm) */			 
						 		 
	num_states = c->num_control_points * 3;
	optimizer_type = gsl_multimin_fdfminimizer_vector_bfgs2;
	optimizer = gsl_multimin_fdfminimizer_alloc(optimizer_type, num_states);
	
	set_cost_function_params(&cost_function_params, c->start_points, c->end_points, c->pre_points, c->post_points, c->num_strands, c->num_control_points, c->num_strand_control_points, c->fov, c->strand_r, sample_density,  cost_term_weights, cost_term_powers, samples_store, grid_reference, grid_reference_size); 
	
	cost_function_struct.n = num_states;
	
	/* This is where our custom cost function ('cost_function' in 'cost_function.c') is passed to the optimisation algorithm */ 
	cost_function_struct.f = &cost_function_f;
	cost_function_struct.df = &cost_function_df;
	cost_function_struct.fdf = &cost_function_fdf;
	cost_function_struct.params = (void *)&cost_function_params;
	
	state_vector = gsl_vector_alloc(num_states);
	
	//printf("State vector stride: %d\n", (int)(state_vector->stride));
	
	/* In order to move the state away from any shallow parts of the cost function (which are not ideal points to start the algorithm from), the state is slightly perturbed (by a gaussian distributed random variable) away from any highly regular configuration it may have started in (i.e. all strands are straight (from 'rand_init') or neighbouring strands are hexagonally packed (from 'subdiv') )*/ 
	
	rand_gen = gsl_rng_alloc(gsl_rng_taus);
	gsl_rng_set(rand_gen, jitter_seed);
		
	for (state_i=0; state_i < num_states; state_i++) {
		state_vector->data[state_i * state_vector->stride] = c->control_points[state_i] + gsl_rng_uniform(rand_gen) * jitter_dev;
	}
	
	gsl_multimin_fdfminimizer_set(optimizer, &cost_function_struct, state_vector, initial_step_size, step_tol);
	
	
	total_time = 0.0;
	
	iteration = 0;
	printf("\n Iteration\t                Func Val.\t               |Gradient|\t             dx Avg.\t              dx Max\t                |dx|\t     Time taken (s)\tNum. cost function calls\n");
	fflush(stdout);
	printf("------------\t-------------------------\t-------------------------\t--------------------\t--------------------\t--------------------\t-------------------\t------------------------\n");
	fflush(stdout);
	printf("            \t                         \t                         \t                    \t                    \t                    \t                   \t");
	fflush(stdout);


	
	
	iterations_next_dirname = malloc(sizeof(char) * (17 + max_iter_label_width));

	
	do {
		
		start = clock();
	
		/* Perform an iteration */
		iteration++;
	/*	num_cost_function_calls = 0;*/
		
		gsl_multimin_fdfminimizer_iterate(optimizer);
		
		status = gsl_multimin_test_gradient(optimizer->gradient, optimization_tol);	
		
		max_dx = 0.0;
		

		/* Determine sum statistics about the iteration */
		dx_sum_sq = 0;
		dx_sum = 0;
		for (state_i=0; state_i < num_states; state_i++) {
			dx = optimizer->x->data[state_i] - c->control_points[state_i];
			c->control_points[state_i] = optimizer->x->data[state_i];
			
			if (dx < 0) {
				dx = -dx;
			}
			
			dx_sum += dx;
			dx_sum_sq += dx * dx;
			
			if (dx > max_dx) {
				max_dx = dx;
			}
		}
		
		dx_av = dx_sum/num_states;
		dx_2norm = sqrt(dx_sum_sq);
		
		stop = clock();
	
		time_for_iteration = (double)(stop - start) / CLOCKS_PER_SEC;
	
		total_time += time_for_iteration;
				
		printf("\n%10d\t%25f\t%25f\t%20g\t%20g\t%20g\t%19g\t", iteration, optimizer->f, gsl_blas_dnrm2 (optimizer->gradient), dx_av, max_dx, dx_2norm, time_for_iteration);
		fflush(stdout);
		
		/* Save the configuration after the iteration */

		if ( iterations_save_freq && (iteration % iterations_save_freq == 0)) {
			sprintf(iterations_next_save_path, "%s%citeration_%0*d", iterations_save_path, DIR_SEP, max_iter_label_width, iteration);
						
									
															
			save_collection(c, iterations_next_save_path,0);
		}

		/* Stop the algorithm if there has been no improvement */
		if (max_dx == 0.0) {
			break;
		}

		
	} while (iteration < max_iterations && status == GSL_CONTINUE);
	
	total_time_min = total_time / 60.0;
	
//	if (status == GSL_SUCCESS) {
//		printf("\n\nMinimum was found successfully after %d iterations in %g seconds (%g min):\n", iteration, total_time, total_time_min);
//	} else { 
//		printf("\n\nDid not find minimum after %d iterations in %g seconds (%g min).\n", iteration, total_time, total_time_min);
//	}
//	
	
	printf("\n\nOptimisation terminated after %d iterations in %g seconds (%g minutes).\n",iteration, total_time, total_time_min); 
	
	fflush(stdout);

	/* copy the output of the optimisation back to the original strand collection pointers to give access to the calling function */

	for (state_i=0; state_i < num_states; state_i++) {
		c->control_points[state_i] = gsl_vector_get(optimizer->x, state_i);
	}



	free(iterations_next_dirname);
	
	gsl_multimin_fdfminimizer_free(optimizer);
	
	free_storage_blocks(samples_store, grid_reference, grid_reference_size);

}


void allocate_storage_blocks(Sample_block **start_block, Reference_block **grid_reference, int grid_reference_size) {
	
	int x, y, z;
	
	*start_block = malloc(sizeof(Sample_block));
			
	*grid_reference = malloc(sizeof(Reference_block)* grid_reference_size * grid_reference_size * grid_reference_size );

	
	init_sample_block(*start_block);
	
	for (z = 0; z < grid_reference_size; z++) {
		for (y = 0; y < grid_reference_size; y++) {
			for (x = 0; x < grid_reference_size; x++) {
				init_reference_block(&(*grid_reference)[grid_reference_size * grid_reference_size * z + grid_reference_size * y + x]);
			}
		}
	}
	
	return;

}


void free_storage_blocks(Sample_block *start_block, Reference_block *grid_reference, int grid_reference_size) {
	
	int x, y, z;
	
	free_sample_blocks(start_block);
	
	for (x = 0; x < grid_reference_size; x++) {
		for (y = 0; y < grid_reference_size; y++) {
			for (z = 0; z < grid_reference_size; z++) {
				free_reference_blocks(&grid_reference[grid_reference_size*grid_reference_size*x + grid_reference_size*y + z]);
			}
		}
	}
	
	free(start_block);
	free(grid_reference);

}


