/*
 *  rand_init.c
 *  Numerical Fibre Generator
 *
 *  Created by Tom Close on 25/06/08.
 *  Copyright 2008 Tom Close.
 *  Distributed under the GNU General Public Licence.
 *
 *
 *
 *  This file is part of 'Numerical Fibre Generator'.
 *
 *  'Numerical Fibre Generator' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  'Numerical Fibre Generator' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'Numerical Fibre Generator'.  If not, see <http://www.gnu.org/licenses/>
 */


#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "strand_collection.h"
#include "shared.h"
#include "rand_init.h"


#define SPHERE_R_DEFAULT 1.0
#define CONTROL_POINT_FREQ_DEFAULT 50.0
#define STRAND_R_LBOUND_DEFAULT 0.075
#define STRAND_R_UBOUND_DEFAULT 0.275
#define STRAND_R_BUFFER_RATIO_DEFAULT 1.1
#define MAX_ATTEMPTS_DEFAULT 1000000


/* Handles the input/output and variable initialisation.  All the 'work' is performed by 'rand_init_collection'*/
int main(int argc, char *argv[]) {


	double sphere_r,control_point_freq, strand_r_lbound, strand_r_ubound, strand_r_buffer_ratio;
	unsigned int seed;
	int max_attempts, error;

	Strand_collection c;
	DIR *dir;
	FILE *file;
	char *path, *param_path, *output_dir_path, trailing_slash[2];
	

	int num_params, num_lines;
	char line[200], key[200];
	double value;


	sphere_r = SPHERE_R_DEFAULT;
	control_point_freq = CONTROL_POINT_FREQ_DEFAULT;
	strand_r_lbound = STRAND_R_LBOUND_DEFAULT;
	strand_r_ubound = STRAND_R_UBOUND_DEFAULT;
	strand_r_buffer_ratio = STRAND_R_BUFFER_RATIO_DEFAULT;
	seed = time(NULL);
	max_attempts = MAX_ATTEMPTS_DEFAULT;
	

	if (argc <= 1 || argc >= 4) {
		printf("\nNumerical Fibre Generator (%s): 'rand_init'\n\nNumber of suplied arguments %d.\n\nUsage: \n\t arg[1]: output directory \n\t arg[2]: parameters file \n\n", VERSION_NUM, (argc-1));

		exit(EXIT_FAILURE);

	} else if (argc == 3) {
	
		param_path = argv[2];
		
		if ((file = fopen(param_path, "r")) == NULL) {
			printf("Error! Could not open parameters file %s\n", param_path);
			exit(EXIT_FAILURE);
		}

		num_params = 0;
		num_lines = 0;
		
		printf("\n\nReading parameters file: %s\n\n", param_path);
		fflush(stdout);
		
		while (fgets(line, 200, file) != NULL) {
			
			if (sscanf(line, "%s %lf", key, &value) != EOF) {
				
				if (!strcmp(key, "sphere_r")) {
					sphere_r = value;
					printf("Read value for sphere_r, %g\n", value);
					
				} else if (!strcmp(key, "control_point_freq")) {
					control_point_freq = value;
					printf("Read value for control_point_freq, %g\n", value);
				
				} else if (!strcmp(key, "strand_r_lbound")) {
					strand_r_lbound = value;
					printf("Read value for strand_r_lbound, %g\n", value);
				
				} else if (!strcmp(key, "strand_r_ubound")) {
					strand_r_ubound = value;
					printf("Read value for strand_r_ubound, %g\n", value);
				
				} else if (!strcmp(key, "strand_r_buffer_ratio")) {
					strand_r_buffer_ratio = value;
					printf("Read value for strand_r_buffer_ratio, %g\n", value);
				
				} else if (!strcmp(key, "seed")) {
				
					if (sscanf(line, "%*s %u", &seed) != 1) {
						printf("Could not read seed value.  Randomly generating...");
					} 
				
					printf("Read value for seed, %d\n", seed);
				
				} else if (!strcmp(key, "max_attempts")) {
					max_attempts = (int)value;
					printf("Read value for max_attempts, %d\n", max_attempts);
				
				} else {
					num_params--;
				}
				fflush(stdout);
				num_params++;
			}
			
			num_lines++;
		}
		
		printf("\nRead in %d (of 7) key/value pairs\n", num_params);
		
	} else {
		printf("\n\nNo parameters file was supplied... using defaults \n\n");
	}
	fflush(stdout);

	output_dir_path = argv[1];

		
	printf("\nGenerating strands...\n");
	fflush(stdout);	
		
	rand_init_collection(&c, sphere_r, control_point_freq, strand_r_lbound, strand_r_ubound, strand_r_buffer_ratio, seed, max_attempts);
		
	printf("\nWriting %d strand files to output directory (arg 1): %s\n", c.num_strands, output_dir_path);
	fflush(stdout);
	
	if (save_collection(&c, output_dir_path, 0)) {
		printf("Could not save collection %s, exiting ...", output_dir_path);
		exit(EXIT_FAILURE);
	}
	
	
	int strand_i;

	for (strand_i = 0; strand_i < c.num_strands; strand_i++) {
	
		printf("Created strand %5d - radius %f - length %f\n", strand_i, c.strand_r[strand_i], dist_between_points(&(c.start_points[strand_i * 3]), &(c.end_points[strand_i * 3])) );
		fflush(stdout);
	}
	
	
	if (output_dir_path[strlen(output_dir_path) - 1] == DIR_SEP) {
		strcpy(trailing_slash, "");
	} else {
		strcpy(trailing_slash, DIR_SEP_S);
	}
	
	path = malloc(sizeof(char) * (strlen(output_dir_path) + 100));	
	sprintf(path, "%s%sparameters%s", output_dir_path, trailing_slash, DIR_SEP_S);
	
	if ((dir = opendir(path)) == NULL) {
		
		if ((error = mkdir(path, MY_PERMS)) > 0) {
			printf("Error! Could not create directory %s, (Error code: %d)!\n", path, error);
			return -1;
		}
	} else {
		closedir(dir);
	}
		
	
	strcat(path, "1_rand_init_param.txt");
	
	if ( (file =fopen(path,"w")) == NULL) {
		printf("\nCould not open file %s.!\n", path);
		exit(EXIT_FAILURE);
	} 

	fprintf(file, "seed %u\n", seed);	
	fprintf(file, "sphere_r %g\n", sphere_r);
	fprintf(file, "control_point_freq %g\n", control_point_freq);
	fprintf(file, "strand_r_lbound %g\n", strand_r_lbound);
	fprintf(file, "strand_r_ubound %g\n", strand_r_ubound);
	fprintf(file, "strand_r_buffer_ratio %g\n", strand_r_buffer_ratio);
	fprintf(file, "max_attempts %d\n", max_attempts);
	
	fclose(file);
		
	exit(EXIT_SUCCESS);
}




/* Randomly initialises a collection of strands by generating pairs of points on a sphere then joining them with evenly spaced control points */
int rand_init_collection(Strand_collection *c, double sphere_r, double control_point_freq, double strand_r_lbound, double strand_r_ubound, double strand_r_buffer_ratio, unsigned int seed, int max_attempts) {

	int ubound_num_strands, ubound_num_control_points, num_attempts;
	int reject;
	
	double overlap_dist;

	int point_i, strand_i, start_of_strand_point_i;
	double segment_disp[3], num_segment_disps;
	double seperation_threshold;
	gsl_rng *rand_gen;
	

	ubound_num_strands = (int)( sphere_r * sphere_r / (strand_r_lbound * strand_r_lbound) + 0.5) * 2; 
	
	ubound_num_control_points = (int)(ubound_num_strands * 2.0 * sphere_r *control_point_freq);
	
	collection_alloc(c, ubound_num_strands, ubound_num_control_points);
	
	
     rand_gen = gsl_rng_alloc(gsl_rng_taus);
	gsl_rng_set(rand_gen, seed);
	
	
	
	/* Randomly generate the 'end-points' (including both start_points and end_points) and radii for the strands to be generated */
	
	c->num_strands = 0;
	c->num_control_points = 0;
	
	num_attempts = 0;
	
	while (num_attempts < max_attempts && c->num_strands < ubound_num_strands) {
		
		c->strand_r[c->num_strands] = gsl_ran_flat(rand_gen, strand_r_lbound, strand_r_ubound);
					
		rand_gen_point_on_sphere(&(c->start_points[c->num_strands * 3]), sphere_r, rand_gen);
		rand_gen_point_on_sphere(&(c->end_points[c->num_strands * 3]), sphere_r, rand_gen);		
		
		
		/* Rejection sampling procedure to favour longer strands and thus better fill the sphere controlior */
		seperation_threshold = gsl_ran_flat(rand_gen, -1.0, 1.0) * sphere_r * sphere_r;
		reject = seperation_threshold < dot_product(&(c->start_points[c->num_strands * 3]), &(c->end_points[c->num_strands * 3]));
		
		if ( !reject ) {
		
			for (strand_i=0; strand_i < c->num_strands; strand_i++) {
				
				/* Ensure that no two 'end points' (both start_points and end_points) are closer than the sum of their respective strands' radii scaled by the 'strand_r_buffer_ratio' */ 
				overlap_dist = strand_r_buffer_ratio * (c->strand_r[c->num_strands] + c->strand_r[strand_i]);
				
				if (	(dist_between_points(&(c->start_points[c->num_strands * 3]), &(c->start_points[strand_i * 3])) < overlap_dist) |
					(dist_between_points(&(c->end_points[c->num_strands * 3]), &(c->end_points[strand_i * 3])) < overlap_dist) |
					(dist_between_points(&(c->start_points[c->num_strands * 3]), &(c->end_points[strand_i * 3])) < overlap_dist) |
					(dist_between_points(&(c->end_points[c->num_strands * 3]), &(c->start_points[strand_i * 3])) < overlap_dist) )
				 {
					reject = 1;
					break;
				}
			}
		}
		
		
		if (reject) {
			num_attempts++;
		} else {
		
			c->num_strand_control_points[c->num_strands] = (int)floor(dist_between_points(&(c->start_points[c->num_strands * 3]), &(c->end_points[c->num_strands * 3])) *control_point_freq); 
			c->num_control_points += c->num_strand_control_points[c->num_strands];
			c->bundle_i_of_strand[strand_i] = strand_i;
			c->num_strands++;
			num_attempts = 0;
		}

	}
	
	
	/* Construct the strand collection including the joining control points from the randomly generated 'end-points' and strand radii. */
	
	point_i = 0;
	
	for (strand_i = 0; strand_i < c->num_strands; strand_i++) {
	
		start_of_strand_point_i = point_i;
		
		segment_disp[X] =  (c->end_points[strand_i * 3 + X] - c->start_points[strand_i * 3 + X]) / (c->num_strand_control_points[strand_i] + 1);
		segment_disp[Y] =  (c->end_points[strand_i * 3 + Y] - c->start_points[strand_i * 3 + Y]) / (c->num_strand_control_points[strand_i] + 1);
		segment_disp[Z] =  (c->end_points[strand_i * 3 + Z] - c->start_points[strand_i * 3 + Z]) / (c->num_strand_control_points[strand_i] + 1);

	
		for ( ; point_i < (start_of_strand_point_i + c->num_strand_control_points[strand_i]); point_i ++) {
		
			num_segment_disps = (double)(point_i - start_of_strand_point_i + 1);
		
			c->control_points[point_i * 3 + X] = c->start_points[strand_i * 3 + X] + num_segment_disps * segment_disp[X];
			c->control_points[point_i * 3 + Y] = c->start_points[strand_i * 3 + Y] + num_segment_disps * segment_disp[Y];
			c->control_points[point_i * 3 + Z] = c->start_points[strand_i * 3 + Z] + num_segment_disps * segment_disp[Z];
		
		}
	
	
		/* Define the reference point for the bending cost of the first segment by extending the start_point/end_point outwards from the sphere.  If you want to use a surface other than a sphere this is how you would define the perpendicular direction.  By defining the pre_point/post_point extended from their start_point/end_point in the perpendicular direction.  */
			
		c->pre_points[strand_i * 3 + X] = c->start_points[strand_i * 3 + X] * (1.0 + STRAND_EXT_FRACT);
		c->pre_points[strand_i * 3 + Y] = c->start_points[strand_i * 3 + Y] * (1.0 + STRAND_EXT_FRACT);
		c->pre_points[strand_i * 3 + Z] = c->start_points[strand_i * 3 + Z] * (1.0 + STRAND_EXT_FRACT);

		c->post_points[strand_i * 3 + X] = c->end_points[strand_i * 3 + X] * (1.0 + STRAND_EXT_FRACT);
		c->post_points[strand_i * 3 + Y] = c->end_points[strand_i * 3 + Y] * (1.0 + STRAND_EXT_FRACT);
		c->post_points[strand_i * 3 + Z] = c->end_points[strand_i * 3 + Z] * (1.0 + STRAND_EXT_FRACT); 
	
		construct_strand(&(c->strands[strand_i]), strand_i, strand_i, &(c->control_points[start_of_strand_point_i * 3]), &(c->start_points[strand_i * 3]), &(c->end_points[strand_i * 3]), &(c->pre_points[strand_i * 3]), &(c->post_points[strand_i * 3]), &(c->segments[start_of_strand_point_i + 3 * strand_i]), c->num_strand_control_points[strand_i], 0.0, c->strand_r[strand_i]); 

	}


	return 0;
}



/* As the name suggests, randomly generates a point on a sphere */
void rand_gen_point_on_sphere(double *point, double sphere_r, const gsl_rng *random_gen) {
	
	double length;

	point[X] = gsl_ran_gaussian(random_gen, 1.0);
	point[Y] = gsl_ran_gaussian(random_gen, 1.0);
	point[Z] = gsl_ran_gaussian(random_gen, 1.0);
	
	length = sqrt(point[X] * point[X] + point[Y] * point[Y] + point[Z] * point[Z]);
	
	point[X] = sphere_r * point[X]/length;
	point[Y] = sphere_r * point[Y]/length;
	point[Z] = sphere_r * point[Z]/length;
	
	
}


