/*
 *  trim.c
 *  Numerical Fibre Generator
 *
 *  Created by Tom Close on 25/06/08.
 *  Copyright 2008 Tom Close.
 *  Distributed under the GNU General Public Licence.
 *
 *
 *
 *  This file is part of 'Numerical Fibre Generator'.
 *
 *  'Numerical Fibre Generator' is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  'Numerical Fibre Generator' is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 'Numerical Fibre Generator'.  If not, see <http://www.gnu.org/licenses/>
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>

#include "strand.h"
#include "strand_collection.h"
#include "segment.h"
#include "shared.h"
#include "strand_section.h"

#include "trim.h"


#define NEW_SPHERE_R_DEFAULT 1.0
#define LENGTH_REJECT_THRESHOLD_DEFAULT 0.2

int main(int argc, char *argv[]) {

	int error;
	double new_sphere_r, length_reject_threshold;

	Strand_collection c, trim_c;
	DIR *dir;	
	FILE *param_file, *file;
	char *param_path, *input_dir_path,  *output_dir_path, *path;
	

	int num_params, num_lines, stage_num;
	char line[200], key[200];
	double value;


	if (argc != 4) {
	
		printf("\nNumerical Fibre Generator (%s): 'trim'\n\nNumber of suplied arguments %d.\n\n Usage: \n\t arg[1]: input directory \n\t arg[2]: output directory \n\t arg[3]: parameters file\n\n", VERSION_NUM, (argc-1));
		exit(EXIT_FAILURE);
	}

	param_path = argv[3];
	
	input_dir_path = argv[1];
	output_dir_path = argv[2];
	
	
	if ((param_file = fopen(param_path, "r")) == NULL) {
		printf("Error! Could not open file %s\n", param_path);
		exit(EXIT_FAILURE);
	}

	if ((dir = opendir(output_dir_path)) == NULL) {
		if ((error = mkdir(output_dir_path, MY_PERMS)) > 0) {
			printf("Error! Could not create directory %s, (Error code: %d)!\n", argv[0], error);
			exit(EXIT_FAILURE);
		}
	} else {
		closedir(dir);
	}	
	
	
	new_sphere_r = NEW_SPHERE_R_DEFAULT;
	length_reject_threshold = LENGTH_REJECT_THRESHOLD_DEFAULT;
			
	num_params = 0;
	num_lines = 0;
	
	printf("\n\nReading parameters file: %s\n\n", param_path);
	fflush(stdout);
	
	
	while (fgets(line, 200, param_file) != NULL) {
		
		if (sscanf(line, "%s %lf", key, &value) != EOF) {
			
			if (!strcmp(key, "new_sphere_r")) {
				new_sphere_r = value;
				printf("Read value for new_sphere_r, %g\n", new_sphere_r);
			
			} else if (!strcmp(key, "length_reject_threshold")) {
				length_reject_threshold = value;
				printf("Read value for length_reject_threshold, %g\n", length_reject_threshold);
			
			} else {
				num_params--;
			}
			num_params++;
		}
		
		num_lines++;
	}
	
	printf("\nRead in %d (of 2) key/value pairs\n", num_params); 
	fflush(stdout);

	
	if (load_collection(&c, input_dir_path)) {
		printf("Could not load strands from directory %s, exiting ...\n", input_dir_path);
		exit(EXIT_FAILURE);
	
	}
						
	printf("\nTrimming strands...\n");	
	fflush(stdout);	
		
	trim(&trim_c, &c, new_sphere_r, length_reject_threshold);
	
	printf("\nWriting %d strand files to: %s\n", trim_c.num_strands, output_dir_path);
	fflush(stdout);
	
	if (save_collection(&trim_c, output_dir_path, 0)) {
		printf("\nCould not save strands to directory %s, exiting ...\n", output_dir_path);
		exit(EXIT_FAILURE);
	
	}
		
		
	/* Copy parameter files to new directory for record of what has been performed */
	
	stage_num = copy_parameters(input_dir_path, output_dir_path);
	if (stage_num != -1) {
		path = malloc(sizeof(char) * (strlen(output_dir_path) + 80));	

		sprintf(path, "%s%cparameters%c%d_trim_param.txt", output_dir_path, DIR_SEP, DIR_SEP, stage_num);
		
		if ( (file =fopen(path,"w")) == NULL) {
			printf("\nCould not open parameters save file %s.!\n", path);
			exit(EXIT_SUCCESS);
		} 
										
		free(path);
																																					
		fprintf(file, "new_sphere_r %g\n", new_sphere_r);
		fprintf(file, "length_reject_threshold %g\n", length_reject_threshold);
		
		
		if (fclose(file)) {
			printf("Error closing parameters file copy\n");
		}
	}	
		
		
	exit(EXIT_SUCCESS);

}


void trim(Strand_collection *trim_c, Strand_collection *c, double new_sphere_r, double length_reject_threshold) {


	int strand_i, num_trimmed_strands, num_trim_control_points, start_of_strand_control_point_i, control_point_i;
	Strand_section **strand_sections, *strand_section;
	Segment *segment;
	Strand *strand;
	int debug;
	
	double *extra_points;
	int num_extra_points;
	
	strand_sections = malloc(sizeof(Strand_section*) * c->num_strands);

	extra_points = malloc(sizeof(double) * (3 * c->num_strands * 2));

	num_trimmed_strands = 0;
	num_trim_control_points = 0;
	num_extra_points = 0;
	
	for (strand_i = 0; strand_i < c->num_strands; strand_i++) {
	

		if (strand_i == 167) {
			printf("original strand_i = %d\n", strand_i);
		}	

	
		strand = &(c->strands[strand_i]);
	
		
		strand_section = strand_sections[strand_i] = NULL;

		segment = strand->pre_segment;
	
		
		if (vector_norm(segment->start_point.pos) < new_sphere_r) {

			strand_section = strand_sections[strand_i] = malloc(sizeof(Strand_section));
			strand_section->prev_section = NULL;
			strand_section_init(strand_section, strand);
			
			strand_section->pre_point = &(extra_points[num_extra_points * 3]);
			num_extra_points++;
			
			strand_section->pre_point[X] = segment->start_point.pos[X] - segment->disp[X];  
			strand_section->pre_point[Y] = segment->start_point.pos[Y] - segment->disp[Y];
			strand_section->pre_point[Z] = segment->start_point.pos[Z] - segment->disp[Z];						
			
			strand_section->entry_point[X] = segment->start_point.pos[X];
			strand_section->entry_point[Y] = segment->start_point.pos[Y];
			strand_section->entry_point[Z] = segment->start_point.pos[Z];
			strand_section->in_sphere = 1;
			strand_section->start_segment = segment->next_segment;
		} 
		
		
		while (segment->next_segment != NULL) {
		
			segment = segment->next_segment;
			if ( ( (strand_section == NULL) || !(strand_section->in_sphere) ) && ( vector_norm(segment->start_point.pos) < new_sphere_r ) ) {
				
				
				if (strand_section == NULL) {
					strand_section = strand_sections[strand_i] = malloc(sizeof(Strand_section));
					strand_section->prev_section = NULL;
					strand_section_init(strand_section, strand);
				
				} else {
					strand_section = strand_section_alloc(strand_section, strand);
				}
				
				set_section_entry(strand_section, segment->prev_segment, new_sphere_r);
				
			} else if ( (strand_section != NULL) && strand_section->in_sphere) {
			 
				if (vector_norm(segment->start_point.pos) > new_sphere_r ) {
				
					set_section_exit(strand_section, segment->prev_segment, new_sphere_r);
				
					if ( section_length(strand_section) >= length_reject_threshold)  {
						strand_section->length_accepted = 1;
						num_trimmed_strands++;
					}
					
				} else {
					strand_section->num_control_points++;
					num_trim_control_points++;
				}
					
			} 
			
		
		}
		
		if ((strand_section != NULL) && strand_section->in_sphere) {
		
			strand_section->exit_point[X] = segment->end_point.pos[X];
			strand_section->exit_point[Y] = segment->end_point.pos[Y];
			strand_section->exit_point[Z] = segment->end_point.pos[Z];
			
			strand_section->post_point = &(extra_points[num_extra_points * 3]);
			num_extra_points++;
			
			strand_section->post_point[X] = segment->end_point.pos[X] + segment->disp[X];  
			strand_section->post_point[Y] = segment->end_point.pos[Y] + segment->disp[Y];
			strand_section->post_point[Z] = segment->end_point.pos[Z] + segment->disp[Z];	
				
		
			if ( section_length(strand_section) >= length_reject_threshold)  {
				strand_section->length_accepted = 1;
				num_trimmed_strands++;
			}
		
		}
	
		
		
	}
	

	collection_alloc(trim_c, num_trimmed_strands, num_trim_control_points);
	trim_c->num_strands = 0;
	trim_c->num_control_points = 0;
	
	trim_c->sphere_r = new_sphere_r;
	trim_c->fov = new_sphere_r;

	for (strand_i = 0; strand_i < c->num_strands; strand_i++) {
	
	
		if (strand_i == 24) {
			debug = 0;
		}
	
		strand_section = strand_sections[strand_i];
	
		while (strand_section != NULL) {
		
			if (strand_section->length_accepted) {
		
				trim_c->start_points[(trim_c->num_strands * 3) + X] = strand_section->entry_point[X]; 
				trim_c->start_points[(trim_c->num_strands * 3) + Y] = strand_section->entry_point[Y];
				trim_c->start_points[(trim_c->num_strands * 3) + Z] = strand_section->entry_point[Z];						
			
				trim_c->end_points[(trim_c->num_strands * 3) + X] = strand_section->exit_point[X]; 
				trim_c->end_points[(trim_c->num_strands * 3) + Y] = strand_section->exit_point[Y];
				trim_c->end_points[(trim_c->num_strands * 3) + Z] = strand_section->exit_point[Z];		
			
			
				if (strand_section->pre_point != NULL) {
					trim_c->pre_points[(trim_c->num_strands * 3) + X] = strand_section->pre_point[X];
					trim_c->pre_points[(trim_c->num_strands * 3) + Y] = strand_section->pre_point[Y];
					trim_c->pre_points[(trim_c->num_strands * 3) + Z] = strand_section->pre_point[Z];
					
				} else {
				
					trim_c->pre_points[(trim_c->num_strands * 3) + X] = strand_section->entry_point[X] * (1.0 + STRAND_EXT_FRACT);
					trim_c->pre_points[(trim_c->num_strands * 3) + Y] = strand_section->entry_point[Y] * (1.0 + STRAND_EXT_FRACT);
					trim_c->pre_points[(trim_c->num_strands * 3) + Z] = strand_section->entry_point[Z] * (1.0 + STRAND_EXT_FRACT);
					
				}
			
			
				if (strand_section->post_point != NULL) {
					trim_c->post_points[(trim_c->num_strands * 3) + X] = strand_section->post_point[X];
					trim_c->post_points[(trim_c->num_strands * 3) + Y] = strand_section->post_point[Y];
					trim_c->post_points[(trim_c->num_strands * 3) + Z] = strand_section->post_point[Z];
					
				} else {
				
					trim_c->post_points[(trim_c->num_strands * 3) + X] = strand_section->exit_point[X] * (1.0 + STRAND_EXT_FRACT);
					trim_c->post_points[(trim_c->num_strands * 3) + Y] = strand_section->exit_point[Y] * (1.0 + STRAND_EXT_FRACT);
					trim_c->post_points[(trim_c->num_strands * 3) + Z] = strand_section->exit_point[Z] * (1.0 + STRAND_EXT_FRACT);
					
				}
			
			
				trim_c->num_strand_control_points[trim_c->num_strands] = strand_section->num_control_points;
				trim_c->strand_r[trim_c->num_strands] = strand_section->strand->radius;
				trim_c->bundle_i_of_strand[trim_c->num_strands] = strand_section->strand->bundle_i;
			
				segment = strand_section->start_segment;
			
				start_of_strand_control_point_i = trim_c->num_control_points;
			
				for (control_point_i = 0; control_point_i < strand_section->num_control_points; control_point_i++) {
					
					trim_c->control_points[trim_c->num_control_points * 3 + X] = segment->start_point.pos[X]; 
					trim_c->control_points[trim_c->num_control_points * 3 + Y] = segment->start_point.pos[Y];
					trim_c->control_points[trim_c->num_control_points * 3 + Z] = segment->start_point.pos[Z];
					
					trim_c->num_control_points++;
					
					segment = segment->next_segment;
				}
				
				debug = start_of_strand_control_point_i + 3 * trim_c->num_strands;
				
				construct_strand(&(trim_c->strands[trim_c->num_strands]), trim_c->num_strands, trim_c->bundle_i_of_strand[trim_c->num_strands], &(trim_c->control_points[start_of_strand_control_point_i]), &(trim_c->start_points[trim_c->num_strands]), &(trim_c->end_points[trim_c->num_strands]), &(trim_c->pre_points[trim_c->num_strands]), &(trim_c->post_points[trim_c->num_strands]), &(trim_c->segments[start_of_strand_control_point_i + 3 * trim_c->num_strands]), trim_c->num_strand_control_points[trim_c->num_strands], 0.0, trim_c->strand_r[trim_c->num_strands]);	
				
				
				trim_c->num_strands++;
					
			}
		
			strand_section = strand_section->next_section;
		}
	
		
			
		free(strand_sections[strand_i]);
	}
	
	free(strand_sections);
	free(extra_points);
}


